list.of.packages <- c("Matrix", "plyr", "multgee")
new.packages <- list.of.packages[!(list.of.packages %in% installed.packages()[,"Package"])]
if(length(new.packages)) install.packages(new.packages)

library(Matrix)
library(MASS)
library(plyr)
library(multgee)


################# functions for calculation of the weights in WGEE and DRGEE #################
gen.weight=function(data=datamiss)
{
#Convert data from long to wide format
widedata=reshape(data,timevar = "time", idvar = "id", v.names=c("Y","Z"),direction = "wide")

#Auxiliary function for creating the weight matrix
fit.psi=function(widedata)
	{
	R1=1*!is.na(widedata$Y.1)
	R2=1*!is.na(widedata$Y.2)
	R3=1*!is.na(widedata$Y.3)
	n=nrow(widedata)
	#

	fit=glm(is.na(X)~Y.1+Z.1,family=binomial,data=widedata)
	mu.x=plogis(fit$coef[1]+fit$coef[2]*widedata$Y.1+fit$coef[3]*widedata$Z.1)
	S.psi.x=cbind(1,widedata$Y.1,widedata$Z.1)*(is.na(widedata$X)-mu.x)
	S.psi.x.vcov=vcov(fit)	
	res.x=rbind(fit$coef); colnames(res.x)=c("psi.0","psi.y","psi.z")
	#
	############################################
	loglik=function(psi)
	{
	psi.01=psi[1];psi.y=psi[2];psi.z=psi[3];psi.r=psi[4]
	#
	R=R1
	cov.z=widedata$Z.2;cov.y=widedata$Y.1;cov.y[is.na(cov.y)]=0
	pr1=1-plogis(psi.01+psi.y*cov.y+psi.r*I(R1==0)+psi.z*cov.z)
	pr0=1-pr1
	m.pred2=cbind(pr0,pr1)
	l2=sum(I(R2==0)*log(m.pred2[,1])+I(R2==1)*log(m.pred2[,2]))
	#
	R=R2
	cov.z=widedata$Z.3;cov.y=widedata$Y.2;cov.y[is.na(cov.y)]=0
	pr1=1-plogis(psi.01+psi.y*cov.y+psi.r*I(R2==0)+psi.z*cov.z)
	pr0=1-pr1
	m.pred3=cbind(pr0,pr1)
	l3=sum(I(R3==0)*log(m.pred3[,1])+I(R3==1)*log(m.pred3[,2]))
	#
	return(-(l2+l3))
	}
	#
	ini=c(0.6,-1.5,-1.3,2.5)
	fit=optim(ini,loglik,method="BFGS",hessian=T)
	res.y=fit$par
	names(res.y)=c("psi.01","psi.y","psi.z","psi.r")
	## Score Function 
	psi=fit$par
	u2=cbind(1,widedata$Y.1,widedata$Z.2,I(R1==0));
	u3=cbind(1,widedata$Y.2,widedata$Z.3,I(R2==0));u3[is.na(u3[,2]),2]=0
	mu2=plogis(u2%*%psi);mu3=plogis(u3%*%psi)
	S.psi.y=u2*as.numeric(I(R2==0)-mu2)+u3*as.numeric(I(R3==0)-mu3)
	S.psi.y.vcov=solve(fit$hessian)
	S.psi=cbind(S.psi.x,S.psi.y)
	S.psi.vcov=bdiag(S.psi.x.vcov,S.psi.y.vcov)
	#
	res=list(res.x=res.x,res.y=res.y)
	return(list(res=res,S.psi=S.psi,S.psi.vcov=S.psi.vcov))
}

##Creates weight matrix
fit.aux=fit.psi(widedata);
S.psi=fit.aux$S.psi;S.psi.vcov=fit.aux$S.psi.vcov
fit.aux=fit.psi(widedata)$res
psi.0=fit.aux$res.x[1];psi.y=fit.aux$res.x[2];psi.z=fit.aux$res.x[3]
pr1=1-plogis(psi.0+psi.y*widedata$Y.1+psi.z*widedata$Z.1)
pr0=1-pr1
m.pred.x=cbind(pr0,pr1)
R.x=1*!is.na(widedata$X);R.y1=1*!is.na(widedata$Y.1);R.y2=1*!is.na(widedata$Y.2);R.y3=1*!is.na(widedata$Y.3)
#
#covariveis
cov.z2=widedata$Z.2;cov.y2=widedata$Y.1;
cov.z3=widedata$Z.3;cov.y3=widedata$Y.2;
cov.y2[is.na(cov.y2)]=0;cov.y3[is.na(cov.y3)]=0;
psi.01=fit.aux$res.y[1];psi.y=fit.aux$res.y[2];psi.z=fit.aux$res.y[3];psi.r=fit.aux$res.y[4]
#############################################
pred.r=function(T,R.p){
	if(T==2) 
	{	
	cov.z=cov.z2;cov.y=cov.y2;
	lp=psi.01+psi.y*cov.y+psi.r*I(R.p==0)+psi.z*cov.z
	}else
	if(T==3) 
	{	
	cov.z=cov.z3;cov.y=cov.y3;
	lp=psi.01+psi.y*cov.y+psi.r*I(R.p==0)+psi.z*cov.z
	}
	pr1=1-plogis(lp)
	pr0=1-pr1
	m.pred=cbind(pr0,pr1)
	return(m.pred)
}
##deltas
#delta11
d.11=R.x/m.pred.x[,2]
#delta22
m.pred2=pred.r(T=2,R.p=1)
d.22=(R.x*R.y2)/(m.pred.x[,2]*m.pred2[,2])
#delta33
a1=pred.r(T=3,R.p=0)[,1]*m.pred2[,1]+pred.r(T=3,R.p=1)[,1]*m.pred2[,2]
a2=pred.r(T=3,R.p=0)[,2]*m.pred2[,1]+pred.r(T=3,R.p=1)[,2]*m.pred2[,2]
m.pred3=cbind(a1,a2)
d.33=(R.x*R.y3)/(m.pred.x[,2]*m.pred3[,2])
#delta12
d.12=d.22
#delta33
d.13=d.33
#delta21
d.21=d.11
#delta23
num=I(R.x==1)*I(R.y2==0)*I(R.y3==1)+I(R.x==1)*I(R.y2==1)*I(R.y3==1)
den=(m.pred.x[,2])*(pred.r(T=3,R.p=0)[,2]*m.pred2[,1]+pred.r(T=3,R.p=1)[,2]*m.pred2[,2])
d.23=num/den
#delta31
d.31=d.11
#delta32
num=I(R.x==1)*I(R.y2==1)*I(R.y3==0)+I(R.x==1)*I(R.y2==1)*I(R.y3==1)
den=(m.pred.x[,2])*(pred.r(T=3,R.p=1)[,1]*m.pred2[,2]+pred.r(T=3,R.p=1)[,2]*m.pred2[,2])
d.32=num/den
###
dados.m2=data
W=cbind(d.11,d.12,d.13,d.21,d.22,d.23,d.31,d.32,d.33)#used in weighting the covariance of the observations
###
pi.12=d.22
pi.13=d.33
pi.23=(I(R.x==1)*I(R.y2==1)*I(R.y3==1))/(m.pred.x[,2]*pred.r(T=3,R.p=1)[,2]*m.pred2[,2])
W.rho=cbind(pi.12,pi.13,pi.23)#used for weighting the Pearson residuals in the correlation structure
###
#############################################
# return(list(S.psi=S.psi,S.psi.vcov=S.psi.vcov,W=W,W.rho=W.rho,m.pred2=m.pred2,m.pred3=m.pred3))
return(list(S.psi=S.psi,S.psi.vcov=S.psi.vcov,W=W,W.rho=W.rho))
}


################# functions for calculation of conditional expectaion in DRGEE #################
gen.cond=function(data=datamiss)
{
patterns=as.data.frame(rbind(##ver uma outra forma de combinao
cbind(y2=rep(c(1,2,3),each=3),y3=rep(c(1,2,3),3),x=0),
cbind(y2=rep(c(1,2,3),each=3),y3=rep(c(1,2,3),3),x=1)))
mm=nrow(patterns)
widedata=reshape(data,timevar = "time", idvar = "id", v.names=c("Y","Z"),direction = "wide")
fit.y2=polr(factor(Y.2)~X+Z.2+Z.1+factor(Y.1),data=widedata)
fit.y3=polr(factor(Y.3)~X+Z.3+Z.2+Z.1+factor(Y.2)+factor(Y.1),data=widedata)
#
fit.x=glm(X~Z.1,data=widedata,family="binomial")
mu.x=plogis(fit.x$coef[1]+fit.x$coef[2]*widedata$Z.1)
S.gamma=cbind(1,widedata$Z.1)*(widedata$X-mu.x)
S.gamma[which(is.na(apply(S.gamma,1,sum))),]=0
S.gamma.vcov=vcov(fit.x)
##Modelo para x
f.x=function(aux.dados,fit.x)
{
	x.obs=aux.dados$X
	xbeta=fit.x$coef[2]*widedata$Z.1
	mu=plogis(fit.x$coef[1]+xbeta)
	prob.x=mu*I(x.obs==1)+(1-mu)*I(x.obs==0)
	return(prob.x)
}
##Modelo para y
f.y2=function(aux.dados,fit.y2)
{
	y.obs=aux.dados$Y.2
	xbeta=fit.y2$coef[1]*aux.dados$X+fit.y2$coef[2]*aux.dados$Z.2+fit.y2$coef[3]*aux.dados$Z.1+
	fit.y2$coef[4]*I(aux.dados$Y.1==2)+fit.y2$coef[5]*I(aux.dados$Y.1==3)
	mu1=plogis(fit.y2$zeta[1]-xbeta)
	mu2=plogis(fit.y2$zeta[2]-xbeta)
	mu=cbind(mu1,mu2-mu1,1-mu2)
	prob.y=mu[,1]*I(y.obs==1)+mu[,2]*I(y.obs==2)+mu[,3]*I(y.obs==3)
	return(prob.y)
}
f.y3=function(aux.dados,fit.y3)
{
	y.obs=aux.dados$Y.3
	xbeta=fit.y3$coef[1]*aux.dados$X+fit.y3$coef[2]*aux.dados$Z.3+fit.y3$coef[3]*aux.dados$Z.2+fit.y3$coef[4]*aux.dados$Z.1+	fit.y3$coef[5]*I(aux.dados$Y.2==2)+fit.y3$coef[6]*I(aux.dados$Y.2==3)+
	fit.y3$coef[7]*I(aux.dados$Y.1==2)+fit.y3$coef[8]*I(aux.dados$Y.1==3)
	mu1=plogis(fit.y3$zeta[1]-xbeta)
	mu2=plogis(fit.y3$zeta[2]-xbeta)
	mu=cbind(mu1,mu2-mu1,1-mu2)
	prob.y=mu[,1]*I(y.obs==1)+mu[,2]*I(y.obs==2)+mu[,3]*I(y.obs==3)
	return(prob.y)
}
##Calculation of wxy weights 
X.imp=data$X
Z.imp=data$Z
Y.imp=data$Y
X.imp[is.na(X.imp)]=0
Z.imp[is.na(Z.imp)]=0
Y.imp[is.na(Y.imp)]=0
D.p=rep(0,nrow(data))
# nomes=names(widedata)
	for(j in 1:mm)
	{
	aux.dados=widedata
	aux.patterns=patterns[j,]
	aux.dados$Y.2=aux.patterns$y2;aux.dados$Y.3=aux.patterns$y3;aux.dados$X=aux.patterns$x
	#
	p.y2=f.y2(aux.dados,fit.y2)
	p.y3=f.y3(aux.dados,fit.y3)
	p.x=f.x(aux.dados,fit.x)
	#
	aux.dados$p.y=(p.y2*p.y3)*(p.x)
	imp.longo=reshape(aux.dados,varying=list(c(3,5,7),c(4,6,8)),v.names=c("Y","Z"),direction="long")
	id.imp=imp.longo$id;t.imp=imp.longo$time
	X.imp=cbind(X.imp,imp.longo$X)
	Z.imp=cbind(Z.imp,imp.longo$Z)
	Y.imp=cbind(Y.imp,imp.longo$Y)	
	D.p=cbind(D.p,imp.longo$p.y)
	}
	D.p[,1]=1
X.imp=cbind(id.imp,t.imp,X.imp);Z.imp=cbind(id.imp,t.imp,Z.imp);Y.imp=cbind(id.imp,t.imp,Y.imp);D.p=cbind(id.imp,t.imp,D.p)
colnames(X.imp)=c("id","time","X",paste("X.imp",1:mm,sep=""))
colnames(Z.imp)=c("id","time","Z",paste("Z.imp",1:mm,sep=""))
colnames(D.p)=c("id","time","D.p",paste("D.p",1:mm,sep=""))
colnames(Y.imp)=c("id","time","y.cres",paste("Y.imp",1:mm,sep=""))
return(list(X=X.imp,Z=Z.imp,Y=Y.imp,D.p=D.p,S.gamma=S.gamma,S.gamma.vcov=S.gamma.vcov))
}


################# functions for calculation of correlation coefficients #################
rho.weight=function(Y1=NULL,mu=NULL,corstr=NULL,W.rho=NULL,data=NULL)
{
n=length(unique(data$id))
m0=matrix(0,2,2)
###Estimation of rho
if(corstr=="exch")
	{
	index2=rep(1:n,each=6)
	rho.est=matrix(0,2,2)
	for(i in 1:n)
		{
		mu.aux=mu[index2==i]
		y.aux=Y1[index2==i]
		mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6];
		y1.aux=y.aux[1:2];y2.aux=y.aux[3:4];y3.aux=y.aux[5:6]
		F1.aux=diag(mu1.aux*(1-mu1.aux));F2.aux=diag(mu2.aux*(1-mu2.aux));F3.aux=diag(mu3.aux*(1-mu3.aux))
		e1=diag(diag(F1.aux)^(-1/2))%*%(y1.aux-mu1.aux)
		e2=diag(diag(F2.aux)^(-1/2))%*%(y2.aux-mu2.aux)
		e3=diag(diag(F3.aux)^(-1/2))%*%(y3.aux-mu3.aux)
		rho.est=rho.est+(e1%*%t(e2))*W.rho[i,1]+(e1%*%t(e3))*W.rho[i,2]+(e2%*%t(e3))*W.rho[i,3]			
		}
	rho.est=rho.est/(3*n-4)
	rho.est=(rho.est+t(rho.est))/2
	rho.est=rbind(cbind(m0,rho.est,rho.est),cbind(rho.est,m0,rho.est),cbind(rho.est,rho.est,m0))
	} else
if(corstr=="unst")
	{
	index2=rep(1:n,each=6)
	rho.est12=matrix(0,2,2);rho.est13=matrix(0,2,2);rho.est23=matrix(0,2,2)
	for(i in 1:n)
		{
		mu.aux=mu[index2==i]
		y.aux=Y1[index2==i]
		mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6];
		y1.aux=y.aux[1:2];y2.aux=y.aux[3:4];y3.aux=y.aux[5:6]
		F1.aux=diag(mu1.aux*(1-mu1.aux));F2.aux=diag(mu2.aux*(1-mu2.aux));F3.aux=diag(mu3.aux*(1-mu3.aux))
		e1=diag(diag(F1.aux)^(-1/2))%*%(y1.aux-mu1.aux)
		e2=diag(diag(F2.aux)^(-1/2))%*%(y2.aux-mu2.aux)
		e3=diag(diag(F3.aux)^(-1/2))%*%(y3.aux-mu3.aux)
		rho.est12=rho.est12+(e1%*%t(e2))*W.rho[i,1]		
		rho.est13=rho.est13+(e1%*%t(e3))*W.rho[i,2]
		rho.est23=rho.est23+(e2%*%t(e3))*W.rho[i,3]	
		}
	rho.est12=rho.est12/(n-4);rho.est21=t(rho.est12)
	rho.est13=rho.est13/(n-4);rho.est31=t(rho.est13)
	rho.est23=rho.est23/(n-4);rho.est32=t(rho.est23)
	rho.est=rbind(cbind(m0,rho.est12,rho.est13),cbind(rho.est21,m0,rho.est23),cbind(rho.est31,rho.est32,m0))
	} else
if(corstr=="ind") {rho.est=matrix(0,6,6)}
##
return(rho.est=rho.est)
}

rho.h=function(Y1=NULL,mu=NULL,corstr=NULL,D.p=NULL,W.rho=NULL,data=NULL)
{
n=length(unique(data$id))
m0=matrix(0,2,2)
###Estimation of rho
if(corstr=="exch")
	{
	for(l in 2:3)
		{		
		index2=rep(1:n,each=6)
		rho.aux=matrix(0,2,2)
		for(i in 1:n)
			{
			mu.aux=mu[[l]][index2==i]
			y.aux=Y1[[l]][index2==i]
			d.paux=D.p[,l][index2==i]
			mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6];
			y1.aux=y.aux[1:2];y2.aux=y.aux[3:4];y3.aux=y.aux[5:6]
			A1.aux=diag(mu1.aux*(1-mu1.aux));A2.aux=diag(mu2.aux*(1-mu2.aux));A3.aux=diag(mu3.aux*(1-mu3.aux))
			e1=diag(diag(A1.aux)^(-1/2))%*%(y1.aux-mu1.aux)
			e2=diag(diag(A2.aux)^(-1/2))%*%(y2.aux-mu2.aux)
			e3=diag(diag(A3.aux)^(-1/2))%*%(y3.aux-mu3.aux)
			rho.aux=rho.aux+(e1%*%t(e2))*d.paux[1]*(1-W.rho[i,1])+(e1%*%t(e3))*d.paux[1]*(1-W.rho[i,2])+(e2%*%t(e3))*d.paux[1]*(1-W.rho[i,3])		
			}
		}
		rho.aux=rho.aux/(3*n-4)
		rho.est=rho.aux
		rho.est=(rho.est+t(rho.est))/2
		rho.est=rbind(cbind(m0,rho.est,rho.est),cbind(rho.est,m0,rho.est),cbind(rho.est,rho.est,m0))
	}else
if(corstr=="unst")
	{
	for(l in 2:3)
		{		
			index2=rep(1:n,each=6)
			rho.aux12=rho.aux13=rho.aux23=matrix(0,2,2)
			for(i in 1:n)
				{
				mu.aux=mu[[l]][index2==i]
				y.aux=Y1[[l]][index2==i]
				d.paux=D.p[,l][index2==i]
				mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6];
				y1.aux=y.aux[1:2];y2.aux=y.aux[3:4];y3.aux=y.aux[5:6]
				A1.aux=diag(mu1.aux*(1-mu1.aux));A2.aux=diag(mu2.aux*(1-mu2.aux));A3.aux=diag(mu3.aux*(1-mu3.aux))
				e1=diag(diag(A1.aux)^(-1/2))%*%(y1.aux-mu1.aux)
				e2=diag(diag(A2.aux)^(-1/2))%*%(y2.aux-mu2.aux)
				e3=diag(diag(A3.aux)^(-1/2))%*%(y3.aux-mu3.aux)
				rho.aux12=rho.aux12+(e1%*%t(e2))*d.paux[1]*(1-W.rho[i,1])
				rho.aux13=rho.aux13+(e1%*%t(e3))*d.paux[1]*(1-W.rho[i,2])
				rho.aux23=rho.aux23+(e2%*%t(e3))*d.paux[1]*(1-W.rho[i,3])						
				}
			rho.est12=rho.aux12/(n-4);rho.est21=t(rho.est12)
			rho.est13=rho.aux13/(n-4);rho.est31=t(rho.est13)
			rho.est23=rho.aux23/(n-4);rho.est32=t(rho.est23)
			rho.est=rbind(cbind(m0,rho.est12,rho.est13),cbind(rho.est21,m0,rho.est23),cbind(rho.est31,rho.est32,m0))
		} 
	}else
if(corstr=="ind") {rho.est=matrix(0,6,6)}
	
return(rho.est=rho.est)
}


################# functions for calculation of covariance matrices #################
gencov.corr.wgee=function(mu=NULL,rho.est=NULL,W=NULL,data=NULL)
{
n=length(unique(data$id))
V.aux=list()
index2=rep(1:n,each=6)
	for(i in 1:n)#comp
	{
		mu.aux=mu[index2==i]
		mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6]
		V1.aux=diag(mu1.aux)-mu1.aux%*%t(mu1.aux);V2.aux=diag(mu2.aux)-mu2.aux%*%t(mu2.aux);
		V3.aux=diag(mu3.aux)-mu3.aux%*%t(mu3.aux)
		F=diag(mu.aux*(1-mu.aux))
		F1.aux=diag(mu1.aux*(1-mu1.aux));F2.aux=diag(mu2.aux*(1-mu2.aux));F3.aux=diag(mu3.aux*(1-mu3.aux));
		C1.aux=diag(diag(F1.aux)^(-1/2))%*%V1.aux%*%diag(diag(F1.aux)^(-1/2))
		C2.aux=diag(diag(F2.aux)^(-1/2))%*%V2.aux%*%diag(diag(F2.aux)^(-1/2))
		C3.aux=diag(diag(F3.aux)^(-1/2))%*%V3.aux%*%diag(diag(F3.aux)^(-1/2))
		C=as.matrix(bdiag(C1.aux,C2.aux,C3.aux)+rho.est)		
		Delta=matrix(W[i,],3,3,byrow=TRUE);Delta=kronecker(Delta,matrix(1,2,2))
		V.aux[[i]]=diag(diag(F^(-1/2)))%*%(ginv(C)*Delta)%*%diag(diag(F^(-1/2)))
	}
	Vinv=as.matrix(bdiag(V.aux))
return(Vinv=Vinv)
}

gencov.odds.wgee=function(mu=NULL,W=NULL,LORterm=NULL,data=NULL)
{
n=length(unique(data$id))
V.aux=list()
index2=rep(1:n,each=6)
	for(i in 1:n)#comp
	{
		mu.aux=mu[index2==i]
		mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6]
		V11=diag(mu1.aux)-mu1.aux%*%t(mu1.aux);V22=diag(mu2.aux)-mu2.aux%*%t(mu2.aux);
		V33=diag(mu3.aux)-mu3.aux%*%t(mu3.aux)
		#timepair=12
		probrow=c(mu1.aux,1-sum(mu1.aux));probcol=c(mu2.aux,1-sum(mu2.aux))
		aux.12=ipfp(LORterm[1,], probrow, probcol,dimension=3)[1:2,1:2]
		#timepair=13
		probrow=c(mu1.aux,1-sum(mu1.aux));probcol=c(mu3.aux,1-sum(mu3.aux))
		aux.13=ipfp(LORterm[2,], probrow, probcol,dimension=3)[1:2,1:2]
		#timepair=23
		probrow=c(mu2.aux,1-sum(mu2.aux));probcol=c(mu3.aux,1-sum(mu3.aux))
		aux.23=ipfp(LORterm[3,], probrow, probcol,dimension=3)[1:2,1:2]
		#
		V12=aux.12-mu1.aux%*%t(mu2.aux);V21=t(V12)
		V13=aux.13-mu1.aux%*%t(mu3.aux);V31=t(V13)
		V23=aux.23-mu2.aux%*%t(mu3.aux);V32=t(V23)
		Delta=matrix(W[i,],3,3,byrow=TRUE);Delta=kronecker(Delta,matrix(1,2,2))
		v.aux=rbind(cbind(V11,V12,V13),cbind(V21,V22,V23),cbind(V31,V32,V33))
		V.aux[[i]]=ginv(v.aux)*Delta
	}
	Vinv=as.matrix(bdiag(V.aux))
return(Vinv=Vinv)
}

gencov.corr.drgee=function(mu=NULL,rho.est=NULL,W=NULL,data=NULL)
{
n=length(unique(data$id))
index2=rep(1:n,each=6)
Vinv=list()
for(l in 1:length(mu))
	{
	V.aux=list()
	for(i in 1:n)#comp
	{
		if(l==1)
		{
			mu.aux=mu[[l]][which(index2==i)]
			mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6]
			V1.aux=diag(mu1.aux)-mu1.aux%*%t(mu1.aux);V2.aux=diag(mu2.aux)-mu2.aux%*%t(mu2.aux);
			V3.aux=diag(mu3.aux)-mu3.aux%*%t(mu3.aux)
			F=diag(mu.aux*(1-mu.aux))
			F1.aux=diag(mu1.aux*(1-mu1.aux));F2.aux=diag(mu2.aux*(1-mu2.aux));F3.aux=diag(mu3.aux*(1-mu3.aux));
			C1.aux=diag(diag(F1.aux)^(-1/2))%*%V1.aux%*%diag(diag(F1.aux)^(-1/2))
			C2.aux=diag(diag(F2.aux)^(-1/2))%*%V2.aux%*%diag(diag(F2.aux)^(-1/2))
			C3.aux=diag(diag(F3.aux)^(-1/2))%*%V3.aux%*%diag(diag(F3.aux)^(-1/2))
			C=as.matrix(bdiag(C1.aux,C2.aux,C3.aux)+rho.est)	
			Delta=matrix(W[i,],3,3,byrow=TRUE);Delta=kronecker(Delta,matrix(1,2,2))
			V.aux[[i]]=diag(diag(F^(-1/2)))%*%(ginv(C)*Delta)%*%diag(diag(F^(-1/2)))
		} else
		if(l>1)
		{
			mu.aux=mu[[l]][which(index2==i)]
			mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6]
			V1.aux=diag(mu1.aux)-mu1.aux%*%t(mu1.aux);V2.aux=diag(mu2.aux)-mu2.aux%*%t(mu2.aux);
			V3.aux=diag(mu3.aux)-mu3.aux%*%t(mu3.aux)
			F=diag(mu.aux*(1-mu.aux))
			F1.aux=diag(mu1.aux*(1-mu1.aux));F2.aux=diag(mu2.aux*(1-mu2.aux));F3.aux=diag(mu3.aux*(1-mu3.aux));
			C1.aux=diag(diag(F1.aux)^(-1/2))%*%V1.aux%*%diag(diag(F1.aux)^(-1/2))
			C2.aux=diag(diag(F2.aux)^(-1/2))%*%V2.aux%*%diag(diag(F2.aux)^(-1/2))
			C3.aux=diag(diag(F3.aux)^(-1/2))%*%V3.aux%*%diag(diag(F3.aux)^(-1/2))
			C=as.matrix(bdiag(C1.aux,C2.aux,C3.aux)+rho.est)		
			Delta=matrix(W[i,],3,3,byrow=TRUE);Delta=kronecker(Delta,matrix(1,2,2))
			V.aux[[i]]=diag(diag(F^(-1/2)))%*%(ginv(C)*(1-Delta))%*%diag(diag(F^(-1/2)))
		}
	}
	Vinv[[l]]=as.matrix(bdiag(V.aux))
	}
return(Vinv)
}

gencov.odds.drgee=function(mu=NULL,W=NULL,LORterm=LORterm,data=NULL)
{
n=length(unique(data$id))
index2=rep(1:n,each=6)
Vinv=list()
for(l in 1:length(mu))
	{
	V.aux=list()
	for(i in 1:n)#comp
	{
		if(l==1)
		{
			mu.aux=mu[[l]][which(index2==i)]
			mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6]
			V11=diag(mu1.aux)-mu1.aux%*%t(mu1.aux);V22=diag(mu2.aux)-mu2.aux%*%t(mu2.aux);
			V33=diag(mu3.aux)-mu3.aux%*%t(mu3.aux)
			# timepair=12
			probrow=c(mu1.aux,1-sum(mu1.aux));probcol=c(mu2.aux,1-sum(mu2.aux))
			aux.12=ipfp(LORterm[1,], probrow, probcol,dimension=3)[1:2,1:2]
			# timepair=13
			probrow=c(mu1.aux,1-sum(mu1.aux));probcol=c(mu3.aux,1-sum(mu3.aux))
			aux.13=ipfp(LORterm[2,], probrow, probcol,dimension=3)[1:2,1:2]
			# timepair=23
			probrow=c(mu2.aux,1-sum(mu2.aux));probcol=c(mu3.aux,1-sum(mu3.aux))
			aux.23=ipfp(LORterm[3,], probrow, probcol,dimension=3)[1:2,1:2]
			
			V12=aux.12-mu1.aux%*%t(mu2.aux);V21=t(V12)
			V13=aux.13-mu1.aux%*%t(mu3.aux);V31=t(V13)
			V23=aux.23-mu2.aux%*%t(mu3.aux);V32=t(V23)
			Delta=matrix(W[i,],3,3,byrow=TRUE);Delta=kronecker(Delta,matrix(1,2,2))
			v.aux=rbind(cbind(V11,V12,V13),cbind(V21,V22,V23),cbind(V31,V32,V33))
			V.aux[[i]]=ginv(v.aux)*Delta
		} else
		if(l>1)
		{
			mu.aux=mu[[l]][which(index2==i)]
			mu1.aux=mu.aux[1:2];mu2.aux=mu.aux[3:4];mu3.aux=mu.aux[5:6]
			V11=diag(mu1.aux)-mu1.aux%*%t(mu1.aux);V22=diag(mu2.aux)-mu2.aux%*%t(mu2.aux);
			V33=diag(mu3.aux)-mu3.aux%*%t(mu3.aux)
			# timepair=12
			probrow=c(mu1.aux,1-sum(mu1.aux));probcol=c(mu2.aux,1-sum(mu2.aux))
			aux.12=ipfp(LORterm[1,], probrow, probcol,dimension=3)[1:2,1:2]
			# timepair=13
			probrow=c(mu1.aux,1-sum(mu1.aux));probcol=c(mu3.aux,1-sum(mu3.aux))
			aux.13=ipfp(LORterm[2,], probrow, probcol,dimension=3)[1:2,1:2]
			# timepair=23
			probrow=c(mu2.aux,1-sum(mu2.aux));probcol=c(mu3.aux,1-sum(mu3.aux))
			aux.23=ipfp(LORterm[3,], probrow, probcol,dimension=3)[1:2,1:2]
			
			V12=aux.12-mu1.aux%*%t(mu2.aux);V21=t(V12)
			V13=aux.13-mu1.aux%*%t(mu3.aux);V31=t(V13)
			V23=aux.23-mu2.aux%*%t(mu3.aux);V32=t(V23)
			Delta=matrix(W[i,],3,3,byrow=TRUE);Delta=kronecker(Delta,matrix(1,2,2))
			v.aux=rbind(cbind(V11,V12,V13),cbind(V21,V22,V23),cbind(V31,V32,V33))
			V.aux[[i]]=ginv(v.aux)*(1-Delta)
		}
	}
	Vinv[[l]]=as.matrix(bdiag(V.aux))
	}
return(Vinv)
}


################# auxiliary functions for calculation of local odds sctructure #################
# these functions are adapted from the work of Touloumis, A. (2015), package 'multgee'
fitLORterm=function(data=NULL,LORstr=NULL)
{
if (LORstr == "RC") LORem <- "2way" else LORem="3way"
timepairs=3
ncategories=3
noccasions=3
noccasionpairs=3
add=1e-4
ncategoriesm1 <- ncategories - 1
data=datacounts(data=data)
if(LORstr=="independence"){LORterm <- matrix(1/ncategories^2, noccasionpairs, ncategories^2)}else
{	
marpars=mmpar(LORem=LORem, LORstr=LORstr, timepairs=timepairs, homogeneous=TRUE)
LORterm=fitmm(data=data, marpars=marpars, homogeneous=TRUE, restricted=FALSE, add=add)  
}
    ans <- diagmod(rep(0, ncategoriesm1 * noccasions))
    k <- 1
    for (i in 1:(noccasions - 1)) {
        for (j in (i + 1):noccasions) {
            ans[seq(ncategoriesm1) + ncategoriesm1 * (i - 1), 
                seq(ncategoriesm1) + ncategoriesm1 * (j - 1)] <- odds.ratio(matrix(LORterm[k, 
                ], ncategories, ncategories))
            k <- k + 1
        }
    }
theta <- ans + t(ans)
return(list(LORterm=LORterm,theta=theta))
}

datacounts=function(data=NULL) 
{
response=data$Y
id=data$id
repeated=data$time
ncategories=max(response,na.rm=TRUE)
    response <- as.numeric(factor(response))
    data <- data.frame(cbind(response, id, repeated))
    data <- reshape(data, v.names = "response", idvar = "id", timevar = "repeated", direction = "wide")
    data <- data[, -1]
    data[is.na(data)] <- 0
    ntimes <- ncol(data)
    notimepairs <- choose(ntimes, 2)
    counts <- rep.int(0, notimepairs * (ncategories^2))
    x <- rep(1:ncategories, each = notimepairs * ncategories)
    y <- rep.int(rep(1:ncategories, each = notimepairs), ncategories)
    tp <- rep.int(1:notimepairs, ncategories^2)
    ind_1 <- 1
    for (categ1 in 1:ncategories) {
        for (categ2 in 1:ncategories) {
            for (ind_2 in 1:(ntimes - 1)) {
                for (ind_3 in (ind_2 + 1):ntimes) {
                  counts[ind_1] <- sum((data[, ind_2] == categ1) & 
                    (data[, ind_3] == categ2))
                  ind_1 <- ind_1 + 1
                }
            }
        }
    }
    data <- data.frame(cbind(counts, x, y, tp))
    data
}

fitmm=function (data, marpars, homogeneous, restricted, add) 
{
    LORstr <- marpars$LORstr
    LORem <- marpars$LORem
    fmla <- marpars$fmla
    ncategories <- max(data$x)
    timepairs <- max(data$tp)
    if (any(data$counts == 0)) 
        data$counts <- data$counts + add
    LORterm <- matrix(0, timepairs, ncategories^2)
    if (LORstr == "uniform" | LORstr == "category.exch") {
        suppressWarnings(fitted.mod <- gnm(fmla, family = poisson, data = data, verbose = FALSE, model = FALSE))
        if (is.null(fitted.mod)) 
            stop("gnm did not converge algorithm")
        coefint <- as.vector(coef(fitted.mod)[pickCoef(fitted.mod,"x:y")])
        if (LORem == "2way") 
            coefint <- mean(coefint)
        LORterm <- matrix(coefint, timepairs, ncategories^2)
        LORterm <- t(apply(LORterm, 1, function(x) exp(x * tcrossprod(1:ncategories))))
    }
    if (LORstr == "time.exch") {
        data$x <- factor(data$x)
        data$y <- factor(data$y)
        if (is.null(restricted)) {
            if (LORem == "3way") {
                if (homogeneous) {
                  suppressWarnings(fitted.mod <- gnm(fmla, family = poisson, 
                    data = data, verbose = FALSE, model = FALSE))
                  if (is.null(fitted.mod)) 
                    stop("gnm did not converge algorithm")
                  coefint <- as.vector(coef(fitted.mod)[pickCoef(fitted.mod, "MultHomog")])
                  coefint <- c(tcrossprod(coefint))
                }
                else {
                  suppressWarnings(fitted.mod <- gnm(fmla, family = poisson, 
                    data = data, verbose = FALSE, model = FALSE))
                  if (is.null(fitted.mod)) 
                    stop("gnm did not converge algorithm")
                  coefint <- as.vector(coef(fitted.mod)[pickCoef(fitted.mod, 
                    "Mult")])
                  coefint <- c(tcrossprod(coefint[-c(1:ncategories)], coefint[1:ncategories]))
                }
                LORterm <- exp(matrix(coefint, nrow = timepairs, ncol = ncategories^2, TRUE))
            }
            else {
                LORterm2 <- LORterm
                for (i in 1:timepairs) {
                  datamar <- data[data$tp == i, ]
                  suppressWarnings(fitted.mod <- gnm(fmla, family = poisson, 
                    data = datamar, verbose = FALSE, model = FALSE))
                  if (homogeneous) {
                    coefint <- as.vector(coef(fitted.mod)[pickCoef(fitted.mod, 
                      "MultHomog")])
                    coefint <- c(tcrossprod(coefint))
                  }
                  else {
                    coefint <- as.vector(coef(fitted.mod)[pickCoef(fitted.mod, 
                      "Mult")])
                    coefint <- c(tcrossprod(coefint[1:ncategories], 
                      coefint[-c(1:ncategories)]))
                  }
                  LORterm2[i, ] <- coefint
                }
                LORterm2 <- colMeans(LORterm2)
                LORterm <- exp(matrix(LORterm2, timepairs, ncategories^2, 
                  TRUE))
            }
        }
        else {
            if (LORem == "3way") {
                if (homogeneous) {
                  coefint <- RRChomog(fmla, data, ncategories)
                }
                else {
                  coefint <- RRCheter(fmla, data, ncategories)
                }
                LORterm <- exp(matrix(coefint, nrow = timepairs, 
                  ncol = ncategories^2, TRUE))
            }
            else {
                LORterm2 <- LORterm
                for (i in 1:timepairs) {
                  datamar <- data[data$tp == i, ]
                  if (homogeneous) {
                    coefint <- RRChomog(fmla, datamar, ncategories)
                  }
                  else {
                    coefint <- RRCheter(fmla, datamar, ncategories)
                  }
                  LORterm2[i, ] <- coefint
                }
                LORterm2 <- colMeans(LORterm2)
                LORterm <- exp(matrix(LORterm2, timepairs, ncategories^2, TRUE))
            }
        }
    }
    if (LORstr == "RC") {
        data$x <- factor(data$x)
        data$y <- factor(data$y)
        for (i in 1:timepairs) {
            datamar <- data[data$tp == i, ]
            suppressWarnings(fitted.mod <- gnm(fmla, family = poisson, 
                data = datamar, verbose = FALSE, model = FALSE))
            if (is.null(restricted)) {
                if (homogeneous) {
                  coefint <- as.vector(coef(fitted.mod)[pickCoef(fitted.mod, 
                    "MultHomog")])
                  coefint <- c(tcrossprod(coefint))
                }
                else {
                  coefint <- as.vector(coef(fitted.mod)[pickCoef(fitted.mod, 
                    "Mult")])
                  coefint <- c(tcrossprod(coefint[1:ncategories], 
                    coefint[-c(1:ncategories)]))
                }
            }
            else {
                coefint <- if (homogeneous) 
                  RRChomog(fmla, datamar, ncategories)
                else RRCheter(fmla, datamar, ncategories)
            }
            LORterm[i, ] <- exp(coefint)
        }
    }
    LORterm <- prop.table(LORterm, 1)
    LORterm
}

diagmod=function (x) 
{
    dimx <- length(x)
    ans <- matrix(0, dimx, dimx)
    ans[1 + 0:(dimx - 1) * (dimx + 1)] <- x
    ans
}

odds.ratio=function (x) 
{
    dimx <- nrow(x)
    x[-1, -1] * x[-dimx, -dimx]/x[-dimx, -1]/x[-1, -dimx]
}

ipfp=function (initial, rowmars, colmars, dimension=3, maxit = 15, tol = 1e-6) 
{
    ans <- initial
    rowsums <- .rowSums(ans, dimension, dimension, FALSE)
    for (i in 1:maxit) {
        ans <- ans * rep.int(rowmars/rowsums, dimension)
        colsums <- .colSums(ans, dimension, dimension, FALSE)
        if (all(abs(colsums - colmars) <= tol)) 
            break
        ans <- ans * rep(colmars/colsums, each = dimension)
        rowsums <- .rowSums(ans, dimension, dimension, FALSE)
        if (all(abs(rowsums - rowmars) <= tol)) 
            break
    }
    matrix(ans, dimension, dimension)
}

mmpar=function (LORem, LORstr, timepairs, homogeneous) 
{
    if (timepairs == 1) {
        LORem <- "2way"
        LORstr <- switch(LORstr, category.exch = "uniform", RC = "time.exch", 
            uniform = "uniform", time.exch = "time.exch")
        if (LORstr == "uniform") 
            fmla <- counts ~ factor(x) + factor(y) + x:y
        if (LORstr == "time.exch") {
            fmla <- if (homogeneous) 
                counts ~ x + y + MultHomog(x, y)
            else counts ~ x + y + Mult(x, y)
        }
    }
    else if (LORem == "2way") {
        if (LORstr == "uniform") 
            fmla <- counts ~ (factor(x) + factor(y)) * factor(tp) + 
                factor(tp):x:y
        if (LORstr == "time.exch" | LORstr == "RC") {
            fmla <- if (homogeneous) 
                counts ~ x + y + MultHomog(x, y)
            else counts ~ x + y + Mult(x, y)
        }
    }
    else {
        if (LORstr == "category.exch") 
            fmla <- counts ~ (factor(x) + factor(y)) * factor(tp) + 
                factor(tp):x:y
        if (LORstr == "uniform") 
            fmla <- counts ~ (factor(x) + factor(y)) * factor(tp) + 
                x:y
        if (LORstr == "time.exch") {
            fmla <- if (homogeneous) 
                counts ~ (x + y) * factor(tp) + MultHomog(x, 
                  y)
            else counts ~ (x + y) * factor(tp) + Mult(x, y)
        }
    }
    list(LORem = LORem, LORstr = LORstr, fmla = fmla)
}

RRChomog=function (fmla, data, ncategories) 
{
    Consmat <- RCconstrains(ncategories, TRUE)
    datax <- factor(data$x)
    datay <- factor(data$y)
    dev <- stop.constrains <- Inf
    noglm <- length(Consmat$nodf[Consmat$nodf < (nlevels(datax) - 
        2)])
    fmla <- update(fmla, ~. - MultHomog(x, y) + MultHomog(z1, 
        z2))
    for (i in 1:noglm) {
        data$z1 <- datax
        data$z2 <- datay
        levels(data$z1) <- levels(data$z2) <- pickcoefind <- Consmat$parscores[i, 
            ]
        suppressWarnings(RRCmod <- gnm(fmla, family = poisson, 
            data = data, verbose = FALSE, model = FALSE))
        if (!is.null(RRCmod)) {
            if (deviance(RRCmod) < dev & RRCmod$conv) {
                pickcoef <- pickCoef(RRCmod, "MultHomog")
                scores <- as.numeric(coef(RRCmod)[pickcoef])[pickcoefind]
                mu <- normscores(scores)
                if (all(diff(mu) >= 0) | all(diff(mu) <= 0)) {
                  dev <- deviance(RRCmod)
                  LORterm <- c(tcrossprod(scores))
                  if (Consmat$nodf[i] < Consmat$nodf[i + 1]) 
                    break
                }
            }
        }
        if (stop.constrains < Consmat$nodf[i + 1]) 
            break
    }
    if (!is.finite(dev)) {
        fmla <- update(fmla, ~. - MultHomog(z1, z2) + x1:x2)
        data$x1 <- as.numeric(datax1)
        data$x2 <- as.numeric(datay1)
        for (i in (noglm + 1):length(Consmat$nodf)) datax1 <- datax
        datay1 <- datay
        levels(datax1) <- levels(datay1) <- pickcoefind <- Consmat$parscores[i, 
            ]
        RRCmod <- suppressWarnings(glm(fmla, data = data, family = poisson, 
            verbose = FALSE, model = FALSE))
        if (deviance(RRCmod) < dev & RRCmod$conv) {
            LORterm <- c(tcrossprod(pickcoefind) * as.numeric(RRCmod$coef["x1:x2"]))
        }
    }
    if (!is.finite(dev)) 
        LORterm <- rep(0, nlevels(datax)^2)
    LORterm
}

RCconstrains=function (ncategories, homogeneous) 
{
    ncategories1 <- ncategories - 1
    nodf <- helpvec <- rep.int(0, ncategories1)
    for (i in 1:(ncategories1 - 1)) {
        helpmat <- t(combn(c(1:ncategories1), i))
        for (j in 1:nrow(helpmat)) {
            helpvec <- rep.int(0, ncategories1)
            helpvec[helpmat[j, ]] <- 1
            nodf <- rbind(nodf, helpvec)
        }
    }
    nodf <- unique(nodf)
    n1 <- nrow(nodf)
    parscores <- matrix(1:ncategories, nrow = n1, ncol = ncategories, 
        byrow = TRUE)
    for (j in 1:ncategories1) {
        parscores[nodf[, j] == 1, j + 1] <- parscores[nodf[, 
            j] == 1, j]
    }
    parscores <- t(apply(parscores, 1, function(x) as.numeric(factor(x))))
    ans <- list(parscores = parscores, nodf = as.numeric(rowSums(nodf)))
    if (!homogeneous) {
        Homogeneous <- ans
        n1 <- length(Homogeneous$nodf)
        parscores <- cbind(apply(Homogeneous$parscores, 2, function(x) rep.int(x, 
            n1)), apply(Homogeneous$parscores, 2, function(x) rep(x, 
            each = n1)))
        nodf <- rep(Homogeneous$nodf, each = n1) + rep.int(Homogeneous$nodf, 
            n1)
        orderedindices <- order(nodf)
        ans <- list(parscores = parscores[orderedindices, ], 
            nodf = nodf[orderedindices])
    }
    ans
}

normscores=function (x) 
(x - mean(x))/sqrt(sum((x - mean(x))^2))







